// perf2rrd: utility to store Nagios performance data into and RRD repository
// Copyright (C) 2004-2012  Marc DeTrano & Gridshield S.A.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// For storing Nagios standard performance data into rrd repository
//

import java.io.*;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.regex.*;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Map;
import java.net.*;

// the following libraries are also free and open software, full copyright notices are included in the jar
// archives.  
import gnu.getopt.*;
import net.stamfest.rrd.*;

public class perf2rrd {
	String MONITORING_CONFIG="";
	String RRD_REPOSITORY="";
	String OBJECT_CACHE="";  //the file where Nagios 2 caches object configuration
        String INTERVAL_FIELD="check_interval";
	String LIVESTATUS_SOCKET="";
        int INT_LENGTH=1;
	boolean DEBUG_MODE=false;
        boolean RUN_ONCE=false;

	int DATA_POINTS=500;     //the number of pixels to have available over a time period
	int DAY_POINTS=86400; //seconds in day
        int [] TIME_PERIODS= {7,31,180,365,550}; //time periods for 1 week, 1 month, and half year--default 
	
	int SLEEP_MILLIS=10000;

	int MONITORING_VERSION=1;
        String MONITORING_BINARY="nagios";
        String MONITORING_ENGINE="nagios";
        
        int TIME_INDEX;  //will hold indexes of output data, as determined from the Nagios perfdata_file_template
        int HOST_INDEX;
        int SVC_INDEX;
        int PERF_INDEX;
        int COUNT_INDEXES=0;
        
        boolean READ_FIFO=false;
        
	//parsing config data for the check_interval value is heavy.  use cache to store values to smooth out creation of many new files.
        Hashtable<String,Double> SERVICE_INTERVAL_CACHE;
        long SERVICE_INTERVAL_CACHE_TIME;
        long SERVICE_INTERVAL_CACHE_OLD=60000; //cache is good for one minute

        RRDp RRD;

	public static void main (String args[]) throws java.io.IOException {
		System.out.println("perf2rrd starting");
		perf2rrd perfParser = new perf2rrd();
		perfParser.startParsing(args);
	}

	void startParsing(String[] args) throws java.io.IOException {
		getOptions(args);
                StringBuffer perfData;
                String perfDataFile="";
                
                System.out.println("Using Monitoring Command: " + MONITORING_BINARY);

                findMonitoringVersion();
		System.out.println("I believe we are using " + MONITORING_ENGINE +" version " + MONITORING_VERSION);

		if (RRD_REPOSITORY.equals("")) {
			RRD_REPOSITORY="/var/log/" + ( MONITORING_ENGINE.equals("nagios") ? "nagios" : (MONITORING_VERSION == 1 ? "icinga" : "icinga2") ) + "/rrd/";
		}

		System.out.println("Using RRD Repository: " + RRD_REPOSITORY);
		if (DEBUG_MODE) {
			System.out.println("Debug Mode is on");
		}
               
		if (MONITORING_CONFIG.equals("")) {
			MONITORING_CONFIG="/etc/" + ( MONITORING_ENGINE.equals("nagios") ? "nagios/nagios.cfg" : (MONITORING_VERSION == 1 ? "icinga/icinga.cfg" : "icinga2/features-available/perf2rrd.conf") );
		}
		System.out.println("Using Monitoring Config: " + MONITORING_CONFIG);


		if ( (MONITORING_ENGINE.equals("nagios")) || ( MONITORING_ENGINE.equals("icinga") && MONITORING_VERSION == 1)) {	
			try {
	                	INT_LENGTH=Integer.parseInt(getNagiosParameter("interval_length"));
                	} catch (Exception ex) {
                    		System.out.println("Trouble parsing interval_length, using 1");
                    		INT_LENGTH=1;
                	}
                	if (INT_LENGTH < 1) {INT_LENGTH=1;}
                

	                if (getNagiosParameter("service_perfdata_file_mode").equals("w") || getNagiosParameter("service_perfdata_file_mode").equals("p")) {
        	            READ_FIFO=true;
                	    System.out.println("Reading perfdata from named pipe.");
	                }
			if (MONITORING_ENGINE.equals("nagios") && MONITORING_VERSION == 1) {
	                	perfDataFile=getNagiosParameter("xpdfile_service_perfdata_file");
        	        } else {
                    		perfDataFile=getNagiosParameter("service_perfdata_file");
			}
		} else {
			perfDataFile=parseSpoolDir();
			perfDataFile=(perfDataFile.equals("") ? "/var/spool/icinga2/perfdata/" : perfDataFile); //Icinga's Default if not set	
		}
                
		File repo = new File(RRD_REPOSITORY);
		repo.mkdir(); //make it if we need it
		if (!(repo.isDirectory())){
			System.out.println("Unable to create RRD Repository");
			System.exit(1);
		}
                
		if (!(new File(perfDataFile).exists())) {
			System.out.println("Unable to find Perf Data File: " + perfDataFile);
			System.exit(1);
		}
		System.out.println("Perf Data File is : " + perfDataFile);
		
		if ( ( MONITORING_ENGINE.equals("icinga") && MONITORING_VERSION == 1)  || ( MONITORING_ENGINE.equals("nagios") && MONITORING_VERSION >= 2)) {
                	OBJECT_CACHE=getNagiosParameter("object_cache_file");

		} else if ( MONITORING_ENGINE.equals("icinga")) {
			// This is fixed for Icinga 2
			OBJECT_CACHE="/var/cache/icinga2/objects.cache";
		}

		if (LIVESTATUS_SOCKET.equals("")) {
			if (!(OBJECT_CACHE.equals(""))) {
				if (!(new File(OBJECT_CACHE).exists())) {
					System.out.println("Cannot find object_cache file :" + OBJECT_CACHE);
					System.exit(1);
				} else {
					System.out.println("Object Cache File is : " + OBJECT_CACHE);
				}
			}
		} else {

			 System.out.println("Using Livestatus Socket: " + LIVESTATUS_SOCKET);
		}
                
                if (MONITORING_ENGINE.equals("nagios") && MONITORING_VERSION <= 2) {
                    INTERVAL_FIELD="normal_check_interval";
                }
                System.out.print("Time periods for RRAs (days): ");
                boolean comma=false;
                for (int d : TIME_PERIODS) {
                    System.out.print((comma ? "," : "") + d);
                    comma=true;
                }
                System.out.println("\nData points per time period: " + DATA_POINTS );
                
		if (MONITORING_ENGINE.equals("icinga")  && MONITORING_VERSION >= 2) {
			INT_LENGTH=60;
		}
                
                if ( DEBUG_MODE ) {System.out.println("Monitoring interval_length: " + INT_LENGTH);}
                if ( DEBUG_MODE ) {System.out.println("Monitoring interval_field name: " + INTERVAL_FIELD);}
                
                
                setIndexes();
                RRD=new RRDp("",null);
		
                //main loop will be slightly different for FIFO reading
                if (READ_FIFO) {
                    while (true) {  //when consumer thread of fifoLoop is terminated, we need to restart the loop
                        if (DEBUG_MODE) {
                            System.out.println("Starting FIFO reading loop");
                        }
                        fifoLoop(perfDataFile);
                    }
                    
                } else {
                	while (true) {  	//main loop
				perfData = ( ( MONITORING_ENGINE.equals("icinga") && MONITORING_VERSION >= 2) ? deSpoolPerfData(perfDataFile) : getPerfData(perfDataFile) );
                            	if ( perfData.length() != 0) {
                                	processPerfData(perfData);
                            	}
                            	if (RUN_ONCE) { break; }
                            	perfData=null;
                            	try {Thread.sleep(SLEEP_MILLIS);} catch (Exception ex) {}
                    	}
                }
	}
        
        void fifoLoop(final String perfDataFile) {

                // For queuing the lines. This object is thread safe by implementation.
                final BlockingQueue<String> queue = new LinkedBlockingQueue<String>();
                final String StopQueue="--stop--";
                // Producer: uses a new thread
                Thread t = new Thread() {
                        public void run() {
                                String line;
                                try {    
                                    BufferedReader input = new BufferedReader(new FileReader(perfDataFile));
                                    try {
                                            while ((line = input.readLine()) != null) {
                                                    queue.put(line);
                                                    if (DEBUG_MODE) {
                                                        System.out.println("Put Queue size is " + queue.size());
                                                    }
                                            }
                                    } catch (IOException ioe) {
                                            System.out.println("Trouble Reading Fifo: " + ioe);
                                    } catch (InterruptedException ie) {
                                            System.out.println("Trouble queueing");
                                    }
                                } catch (IOException fnfe) {
                                    System.out.println("Trouble Reading Fifo: " + fnfe);
                                }
                                if (DEBUG_MODE) {
                                    System.out.println("Thread is exiting");
                                }
                                //signal through the queue
                                try {
                                    queue.put(StopQueue);
                                } catch (InterruptedException ie) {
                                            System.out.println("Trouble queueing");
                                }

                        }
                };
                t.start();
                
                // Consumer: uses the main thread
                // will exit when thread is terminated
                try {
                        while (t.getState() != Thread.State.TERMINATED) {
                                String line = queue.take().toString();
                                if ( line.equals(StopQueue) ) {
                                    break;
                                }
                                if (DEBUG_MODE) {
                                    System.out.println("Take Queue size is " + queue.size());
                                }
                                processPerfData(line);
                        }
                } catch (InterruptedException ie) {
                        System.out.println("Trouble queueing");
                }
        }

	void getOptions(String[] args) {
                Getopt g = new Getopt("perf2rrd", args, "c:n:d:t:p:l:xo?h");
                int o, val=0;
                String arg;
                while ((o = g.getopt()) != -1) {
                        arg=g.getOptarg();
                        try { val=Integer.parseInt(arg); } catch (Exception ex) {}
                        switch(o) {

                                case 'c':
                                        MONITORING_CONFIG=arg;
                                        break;
                                case 'n':
                                        MONITORING_BINARY=arg;
                                        break;          
                                case 'd':
                                        RRD_REPOSITORY=arg;
                                        break;
				case 'x':
                                        DEBUG_MODE=true;
                                        break;
                                case 'o':
                                        RUN_ONCE=true;
                                        break;
                                case 'p':
                                        DATA_POINTS=val;
                                        break;
                                case 't': //parse time periods, list of muliples of one day
                                        String[] periods = arg.split(",") ;
                                        TIME_PERIODS=new int[periods.length];
                                        int i=0;
                                        try {
                                            for (String p : periods ) {
                                                TIME_PERIODS[i]=Integer.valueOf(p);
                                                i++;
                                            }
                                        } catch (NumberFormatException ex) {
                                            System.out.println("Bad value for timeperiod\nSee usage documentation");
                                            System.exit(1);
                                        }
                                        break;
				case 'l':
					LIVESTATUS_SOCKET=arg;
					break;
                                default:
                                        System.out.println("\nUsage:\n");
                                        System.out.println("\t-n Monitoring Command (default 'nagios')");
                                        System.out.println("\t   Note: This command will be run to determine engine (nagios or icinga) and version.");
                                        System.out.println("\t     Other defaults will be set accordingly.");
                                        System.out.println("\t-c Monitoring Config File");
                                        System.out.println("\t    Defaults:");
                                        System.out.println("\t    \tnagios: /etc/nagios/nagios.cfg");
                                        System.out.println("\t    \ticinga: /etc/icinga/icinga.cfg");
                                        System.out.println("\t    \ticinga2: /etc/icinga2/features-available/perf2rrd.conf");
                                        System.out.println("\t-d RRD Repository Dir (default /var/log/{nagios|icinga|icinga2}/rrd)");
                                        System.out.println("\t-x Debug mode on");
                                        System.out.println("\t-o Process data file once then exit");
                                        System.out.println("\t-p Data points per time period (default 500)");
                                        System.out.println("\t-t Comma separated list of time periods for RRAs, given as days.  (default 7,31,180,365,550)");
                                        System.out.println("\t   Note: 24 hours worth of unconsolidated data points are always maintained in one RRA");
					System.out.println("\t-l Livestatus TCP socket {IP}:{port}.  If set, Livestatus will be used instead of the objects.cache file to determine check intervals");
                                        System.out.println("\n");
					System.exit(0);
                        }
                }
        }
        
        void findMonitoringVersion() {
	    // Set globals according to engine and version as deternined from binary
            try {
                String line=null;
                Process vercheck = Runtime.getRuntime().exec(MONITORING_BINARY + " --version");
                BufferedReader verout = new BufferedReader( new InputStreamReader(vercheck.getInputStream()));
                Pattern nagVer=Pattern.compile("^(Nagios|[Ii]cinga)\\D*(\\d*).*");
                while ((line = verout.readLine()) != null) {
                    Matcher m=nagVer.matcher(line);
                    if (m.find()) {
			MONITORING_ENGINE=m.group(1).toLowerCase();
			MONITORING_VERSION=(Integer.parseInt(m.group(2)));
                        return;
                    }
                }
                
            } catch (Exception ex) {
                System.out.println("Problem reading output from command " + MONITORING_BINARY + "\n");
                System.out.println("You may need to set the -n flag\n");
                System.exit(0);
            }
            
            return;
        }

	String getNagiosParameter(String sToken) {
		String retval="", line;

		StringBuffer config = parseConfig(MONITORING_CONFIG);
		BufferedReader br = new BufferedReader(new StringReader(config.toString()));
		try {
			while (((line = br.readLine()) != null) && (retval=="")){
				StringTokenizer tokens = new StringTokenizer(line,"=");
				while (tokens.hasMoreTokens()) {
					if (tokens.nextToken().trim().equals(sToken)) {
						retval=tokens.nextToken().trim();
						break;
					}
				}
			}
		} catch (Exception ex) {
			System.out.println(ex);
		}
		return(retval);
	}

	StringBuffer parseConfig(String nagConfig) {
		//note, this could be called recursively, if Nagios cfg_file or cfg_dir is used

		StringBuffer retval = new StringBuffer();
		try {
			FileReader config = new FileReader(nagConfig);
			BufferedReader br = new BufferedReader(config);
			String line;
			while ( (line = br.readLine()) != null) {
				if (( line.startsWith("#")) || (line.trim().equals(""))){
					continue;
				}
				StringTokenizer tokens = new StringTokenizer(line, "=");
				while (tokens.hasMoreTokens()) {
					String tkn = tokens.nextToken().trim();
					if (tkn.equals("cfg_file")){
						retval.append(parseConfig(tokens.nextToken().trim()));
					}
					if (tkn.equals("cfg_dir")) {
						String[] cfgs = getConfigs(tokens.nextToken().trim());
						for (int i=0; i < cfgs.length; i++) {
							retval.append(parseConfig(cfgs[i]));
						}
					}

				}
				retval.append(line + "\n");
			}
  		} catch (Exception ex) {
			System.out.println("Problem Parsing Config File: " + ex);
		}

		return(retval);

	}

	String parseSpoolDir() {
		// Scans the config for the PerfdataWriter in Icinga2 for the spool directory location
		// This will account for comments, but not much else.  Assumes only "our" writer is defined in the file.
		String retval="";
		try ( 
			FileReader config = new FileReader(MONITORING_CONFIG);
                	BufferedReader br = new BufferedReader(config);
		) {	
                	String line;
			String dirPattern = ".*service_perfdata_path\\s*=\\s*\"([^\"]*?)\".*";
			String cCommentStart = "(.*?)/\\*.*";
			String cCommentEnd = ".*?\\*/(.*)";
			String cInline = "/\\*.*?\\*/";
			String hashComment = "(.*?)#.*";
			Boolean inComment = false;

                	while ( (line = br.readLine()) != null) {
				String stripLine=line;
				if (!inComment) {
					stripLine=stripLine.replaceAll(hashComment,"$1"); //first hash comments to
					stripLine=stripLine.replaceAll(cInline,""); //then single line c comments go
					if (stripLine.matches(cCommentStart)) { //multiline c comment starts
						inComment=true;
						stripLine=stripLine.replaceAll(cCommentStart, "$1");
					}
				} else {
					if (stripLine.matches(cCommentEnd)) {  //multiline c comment ends
						inComment=false;
						stripLine=stripLine.replaceAll(cCommentEnd, "$1");
						stripLine=stripLine.replaceAll(hashComment,"$1");  //then remaining hash comments go
						stripLine=stripLine.replaceAll(cInline,"");  //then remaining incline c comments go
						if (stripLine.matches(cCommentStart)) {  // a remaining multiline start is then still possible
                	                                inComment=true;
        	                                        stripLine=stripLine.replaceAll(cCommentStart, "$1");
	                                        }
					} else {
						stripLine="";
					}
				}

				if (stripLine.matches(dirPattern)) {
					retval=new File(stripLine.replaceAll(dirPattern,"$1")).getParent();

				}
			}
			config.close();

		} catch (Exception ex) {
				System.out.println("Problem reading Config File: " + MONITORING_CONFIG + ex);
		}
		return retval;

	}

	String[] getConfigs(String cfgdir) {
		//list of .cfg files in cfgdir
		ArrayList<String> templist = new ArrayList<String>();
		File[] files = new File(cfgdir).listFiles();
		int count=0;

		for (int i=0;i < files.length; i++) {
			String name = files[i].toString();
			if (files[i].isDirectory()) {   //Nagios 2 recurses the cfg_dirs
				String[] sublist = getConfigs(name);
				for (int j = 0; j < sublist.length; j++) {
					templist.add(sublist[j]);
				}
			}
			if (name.endsWith(".cfg")) {
				templist.add(name);
			}
		}
		String[] retval = (String []) templist.toArray (new String [templist.size ()]);
		return(retval);
	}

        void setIndexes() {

		if ( MONITORING_ENGINE.equals("icinga") && MONITORING_VERSION >=2 ) {
			// Indexes are fixed in this case, use PerfdataWriter Config to set template
			TIME_INDEX = 1;
			HOST_INDEX = 2;
			SVC_INDEX = 3;
			PERF_INDEX = 4;
			COUNT_INDEXES = 5;
			return;
		}

            	String tmpl="", tkn;
            	if ( MONITORING_ENGINE.equals("nagios") && MONITORING_VERSION == 1) {
            	    tmpl=getNagiosParameter("xpdfile_service_perfdata_template");
            	} else {
            	    tmpl=getNagiosParameter("service_perfdata_file_template");
            	}
            	StringTokenizer tokens = new StringTokenizer(tmpl, "\\t");
            	COUNT_INDEXES = tokens.countTokens();
            	for (int i=0; i < COUNT_INDEXES ; i++ ) {
            	    tkn = tokens.nextToken();
            	    if (tkn.equals("$TIMET$")) {
            	        TIME_INDEX = i;
            	    }
            	    if (tkn.equals("$HOSTNAME$")) {
            	        HOST_INDEX = i;
            	    }
            	    if (tkn.equals("$SERVICEDESC$")) {
            	        SVC_INDEX = i;
            	    }
            	    if ( tkn.equals("$SERVICEPERFDATA$") || tkn.equals("$PERFDATA$")) {
            	        PERF_INDEX = i;
            	    }
            	    
            	}
            
        } 
        
	StringBuffer getPerfData(String filename) {
		StringBuffer retval= new StringBuffer("");
		String line;
                
                try {
                    RandomAccessFile perfFile = new RandomAccessFile(filename,"rws");
                    while ((line=perfFile.readLine()) != null) {
                        if (line != "") {
                            retval.append(line + "\n");
                        }
                    }
                    perfFile.setLength(0);
                    perfFile.close();
                } catch (Exception ex) {
                    System.out.println("Trouble Reading Perfdata");
                }

		return retval;
	}
	
	StringBuffer deSpoolPerfData(String filename) {
		StringBuffer retval= new StringBuffer("");
		File spoolDir=new File(filename);
		File[] datafiles = spoolDir.listFiles();

		for (File datafile : datafiles) {
			String datafilePath = filename+"/"+datafile.getName();
			if (DEBUG_MODE) {System.out.println("Reading File: " + datafilePath); }
			retval.append(getPerfData(datafilePath));
			if (!(datafile.delete())) { System.out.println("Warning: could not delete datafile from Spool Directory : " + datafilePath); }
		}
		return retval;
	}
        
        
        void processPerfData(StringBuffer perfData) {
            String line="";
            BufferedReader br = new BufferedReader(new StringReader(perfData.toString()));
            try {
                while ((line = br.readLine()) != null){
                    processPerfData(line);
                }
            } catch (Exception ex) {
                System.out.println("Problem reading perfData StringBuffer");
            }
            
        }
        
	void processPerfData(String line) {
		String host = "", service = "", perfOut ="";
                Long date = new Long(0);
		
                if (DEBUG_MODE) {
                    System.out.println("processPerfData called with: " + line);
                    
                }
                
                StringTokenizer tokens = new StringTokenizer(line,"\t");

                if (tokens.countTokens() == COUNT_INDEXES) {
                        for (int i=0 ; i < COUNT_INDEXES ; i++) {
                            String tkn = tokens.nextToken();
                            if ( i == TIME_INDEX) { date = Long.valueOf(tkn); }
                            if ( i == HOST_INDEX) { host = tkn; }
                            if ( i == SVC_INDEX) { service = tkn; }
                            if ( i == PERF_INDEX) { perfOut = tkn;  }
                        }

                        PerfItem[] perfItems = getItems(perfOut);
                        for (int i=0; i < perfItems.length ; i++){
                                String rrdfile = getRRD(host, service, perfItems[i], date.longValue());//check for file
                                if (!(rrdfile.equals(""))) {
                                        String[] updateCMD = {"update " + rrdfile + " " + date + ":" + perfItems[i].getItemValue()};
                                        try {
                                                RRD.command(updateCMD);
                                                if (DEBUG_MODE) {
                                                        System.out.println("called update with: " + updateCMD[0]);
                                                }
                                        } catch (Exception ex) {
                                                System.out.println("Problem updating: " + rrdfile + ": " + ex);
                                        }
                                }
                        }

                }


	}

	PerfItem[] getItems(String s) {
		ArrayList<PerfItem> tempList = new ArrayList<PerfItem>();
                //replace spaces contained in single quotes with an underscore
                boolean inQuotes=false;
                String stemp="";
                char ctemp;
                if ( s.indexOf('\'') != -1 ) {
                    for ( int i=0; i < s.length(); i++ ) {
                        ctemp=s.charAt(i);
                        if (ctemp == '\'' ) {
                            inQuotes = !(inQuotes);
                            continue;
                        }
                        if ( (inQuotes) && (ctemp == ' ') ) {
                            stemp=stemp + '_';
                        } else {
                            stemp = stemp + ctemp;
                        }
                    }
                    s=stemp;
                }
                
		StringTokenizer tokens = new StringTokenizer(s," ");
		while (tokens.hasMoreTokens()) {
			String temp = tokens.nextToken();
			tempList.add(new PerfItem(temp));
		}
		PerfItem[] retval = (PerfItem []) tempList.toArray (new PerfItem [tempList.size()]);
		return(retval);
	}

	String getRRD(String h, String s, PerfItem p, long d) {
		//return name of RRD, create new one if necessary
		String fname=RRD_REPOSITORY+"/";

		String hostFilename=stringFix(h);
		String serviceFilename=stringFix(s);
		String perfFilename=stringFix(p.getItemName());

		fname=fname + hostFilename+"+"+serviceFilename+"+"+perfFilename+".rrd";
		if (!(new File(fname).exists())) {
			if (!(createRRD(h,s,p,d, fname))) {
				return("");
			}
		}

		return(fname);
	}

	boolean createRRD(String h, String s, PerfItem p, long d, String fname) {
                ArrayList<String> createCMD = new ArrayList<String>();
		String[] RRAs = {"AVERAGE", "MIN", "MAX"};
                int steps;
		int storePoints = (int)(DATA_POINTS * 1.25); //store extra, avoid blocky graphs

		double temp_interval=-1.0;
                p.parseFull();  //we need all the details to create an RRD
		if (h.equals("") || s.equals("") || (p.getItemName()).equals("")) {
			return(false);
		}
		try {
			temp_interval = getServiceInterval(h,s);
		} catch (Exception ex) {
			System.out.println("Trouble Parsing interval- host:" + h + " service: " + s +" error: " + ex);
		}

		if (temp_interval < 0.0) {
			return(false);
		}
           
                temp_interval=temp_interval * INT_LENGTH; //Nagios interval in real seconds
                
                int interval=(int)temp_interval;
                
		int heartbeat=interval * 2;

		String rrdType="GAUGE";
		int type=p.getItemType();
		if (type == PerfItem.COUNTER) {
			rrdType="COUNTER";
		}

		String rrdMin="U";
		String rrdMax="U";
		Double min=p.getItemMin();
		Double max=p.getItemMax();

		if (!(min == null)) {
			rrdMin=min.toString();
		}

		if (!(max == null)) {
			rrdMax=max.toString();
		}

		createCMD.add("create " + fname + " --step " + interval + " --start " + (d-interval));
		createCMD.add("DS:val:" + rrdType + ":" + heartbeat + ":" + rrdMin + ":" + rrdMax);
		for (int i=0; i < RRAs.length ; i++) {
                        //always store all points for 24 hours
                        createCMD.add("RRA:" + RRAs[i] + ":0.5:1:"+ (DAY_POINTS / interval));
			for (int j=0; j < TIME_PERIODS.length; j++) {
				steps=(int)((TIME_PERIODS[j] * DAY_POINTS) / (interval * DATA_POINTS));
                                steps=( steps < 1 ? 1 : steps);
				createCMD.add("RRA:" + RRAs[i] + ":0.5:" + steps + ":" + storePoints);
			}
		}

		try {
                        RRD.command(createCMD.toArray(new String[createCMD.size()]));
		} catch (Exception ex) {
			System.out.println("Problem creating file: " +fname + " : " + ex);
			return(false);
		}

		System.out.println(fname + " created.");
		return(true);

	}

        double getServiceInterval(String h, String s) {
        	//return normal check interval for the object, use cached information if possible
        	// A service name of "_" is interpreted as a host check
            
            //caching is not implemented for ver.1, harder to check for updates in config
            
            if ( MONITORING_ENGINE.equals("nagios") && MONITORING_VERSION == 1 ) {
                        Hashtable services = new Hashtable();
                        Hashtable hosts = parseNagObjects("Host");
                        Hashtable hostgroups = parseHostGroups(hosts);
                        services = parseServices(hosts, hostgroups);
			
			if (s.equals("_")) {
				return (Double.parseDouble(getObjDirective(hosts, h , INTERVAL_FIELD))); 
			} else {
	                        return (Double.parseDouble(getObjDirective(services, h + "+" + s, INTERVAL_FIELD)));
			}
            }
           
		// if there is no cache, the object is not in the cache, or the cache is expired, refresh the cache
            if ( SERVICE_INTERVAL_CACHE == null || (SERVICE_INTERVAL_CACHE.get(h + "+" +s) == null) || (( System.currentTimeMillis() - SERVICE_INTERVAL_CACHE_TIME) > SERVICE_INTERVAL_CACHE_OLD ) ) {
                        System.out.println("Updating Interval Cache.");
                        
                        SERVICE_INTERVAL_CACHE = new Hashtable<String,Double>();
			Hashtable<String,Hashtable<String,String>> services;
			if (LIVESTATUS_SOCKET.equals("")) {
	                        services = parseObjectsCache();
			} else {
				services = getLivestatusObjects();
			}

                        for (Map.Entry<String,Hashtable<String,String>> entry : services.entrySet() ) {
                                SERVICE_INTERVAL_CACHE.put( entry.getKey() , Double.parseDouble((entry.getValue()).get(INTERVAL_FIELD)));
                        }
                        SERVICE_INTERVAL_CACHE_TIME = System.currentTimeMillis();
                        
            }
            
            return ((SERVICE_INTERVAL_CACHE.get(h + "+" +s)));


        }

	Hashtable parseHostGroups(Hashtable hosts) {
		//a host group will have a String index (the name), and a Vector object holding all members of the group.
		//TODO, accomodate hostgroup directive in new Nagios host object configuration
		// and allow for multiple "members" directives in a hostgroup.
		Hashtable<String,Vector> hostgroups = new Hashtable<String,Vector>();
		Hashtable<String,Hashtable<String,String>> hg_config = parseNagObjects("hostgroup");

                for (String currKey : hg_config.keySet()) {
                	String s = getObjDirective(hg_config, currKey, "members");
			StringTokenizer members = new StringTokenizer(s, ",");
			Vector<String> membersList = new Vector<String>();
			while (members.hasMoreTokens()) {
				membersList.add((members.nextToken()).trim());
			}
			hostgroups.put(currKey, membersList);

		}

		return (hostgroups);
	}

	Hashtable parseServices(Hashtable hosts, Hashtable hostgroups) {
		//services, if not indexed by the 'name' directive, will be indexed with host_name+service_description
		Hashtable<String, Hashtable> services = new Hashtable<String, Hashtable>();

		Hashtable<String, Hashtable<String,String>> svc_cfg = parseNagObjects("service");
		
                for (String currKey : svc_cfg.keySet()) {
			if (!(currKey.startsWith("perf2rrd+"))) {  //it has a 'name', so make this entry
				services.put(currKey,svc_cfg.get(currKey));
				//continue;
			}
			String sname = getObjDirective(svc_cfg, currKey ,"service_description");
			String hosts_cfg = getObjDirective(svc_cfg, currKey ,"host_name");
			if (!(hosts_cfg.equals(""))) {
				StringTokenizer hostList = new StringTokenizer (hosts_cfg, ",");
				while (hostList.hasMoreTokens()) {
					services.put( ((String) hostList.nextToken()).trim() + "+" + sname, svc_cfg.get(currKey));
				}
			} else {
				String hostgroupList = getObjDirective(svc_cfg, currKey ,"hostgroup_name");
				if (!(hostgroupList.equals(""))) {
					StringTokenizer groupList = new StringTokenizer (hostgroupList, ",");
					while (groupList.hasMoreTokens()) {
						Vector hlist = (Vector) hostgroups.get(groupList.nextToken().trim());
						Enumeration e = hlist.elements();
						while (e.hasMoreElements()) {
							services.put( ((String) e.nextElement()).trim() + "+" + sname, svc_cfg.get(currKey));
						}
					}
				}
			}
		}
		return(services);
	}
	
	Hashtable<String,Hashtable<String,String>> parseObjectsCache () {
		//with Nagios 2 I have the object_cache file, makes this part much easier
		Hashtable<String, Hashtable<String,String>> services = new Hashtable<String,Hashtable<String,String>>();
		Hashtable<String,Hashtable<String,String>> svc_cfg = parseNagObjects("service"); //this function can handle Nagios 2 via object cache file
		Hashtable<String,Hashtable<String,String>> host_cfg = parseNagObjects("host"); //this function can handle Nagios 2 via object cache file

                for (String currKey : svc_cfg.keySet()) {
			String sname = getObjDirective(svc_cfg, currKey ,"service_description");
			String hostname = getObjDirective(svc_cfg, currKey ,"host_name");  //when using object_cache, I know this will yield one host
			services.put( hostname + "+" + sname, svc_cfg.get(currKey));
		}

		// Adding support of host checks, treated as a 'service'
                for (String currKey : host_cfg.keySet()) {
			String sname = "_";
			String hostname = getObjDirective(host_cfg, currKey ,"host_name");
			services.put( hostname + "+" + sname, host_cfg.get(currKey));
		}

		return(services);
	}

	// Livestatus version of parse Objects, only returns necessary fields
	Hashtable<String,Hashtable<String,String>> getLivestatusObjects () {
		Hashtable<String, Hashtable<String,String>> statusObjects = new Hashtable<String,Hashtable<String,String>>();
		
		ArrayList<String[]> hosts = liveQuery("GET hosts\nColumns: host_name " + INTERVAL_FIELD);
		ArrayList<String[]> services = liveQuery("GET services\nColumns: host_name service_description " + INTERVAL_FIELD);

		for (String[] host : hosts) {
			Hashtable<String,String> thisObj = new Hashtable<String,String>();
			thisObj.put("host_name", host[0]);
			thisObj.put(INTERVAL_FIELD, host[1]);
			statusObjects.put(host[0]+"+_", thisObj);
		}
		for (String[] service : services) {
			Hashtable<String,String> thisObj = new Hashtable<String,String>();
			thisObj.put("host_name", service[0]);
			thisObj.put("service_descripiton", service[1]);
			thisObj.put(INTERVAL_FIELD, service[2]);
			statusObjects.put(service[0] + "+" + service[1], thisObj);
		}

		return(statusObjects);

	}

	ArrayList<String[]> liveQuery(String query) {
		ArrayList<String[]> result = new  ArrayList<String[]>();

		String[] splitAddr = LIVESTATUS_SOCKET.split(":");
                int port = Integer.parseInt(splitAddr[1]);
                String socket_addr = splitAddr[0];
                if (DEBUG_MODE) { System.out.println("Querying Livestatus on: " + socket_addr +" port: " + port); }
                try (
                        Socket live = new Socket(socket_addr, port);
                        PrintWriter out = new PrintWriter(live.getOutputStream(), true);
                        BufferedReader in = new BufferedReader( new InputStreamReader(live.getInputStream()));
                ) {
                        String lineIn;
                        out.println(query+"\n\n");
                        while ( (lineIn = in.readLine()) != null ) {
                                String[] fields=lineIn.split(";");
				result.add(fields);
                        };
		} catch (Exception ex) {
                        System.out.println("Problem reading from Livestatus Socket: " + LIVESTATUS_SOCKET + " Err: " + ex);
                }
		return result;

	}


	String getObjDirective(Hashtable allObjects, Object key, String dName) {
		//key is a String object, used as a key for an object in allObjects
		String retval="";
		Hashtable currObj=(Hashtable) allObjects.get(key);
		if (currObj.containsKey(dName)) {
			retval=(String)currObj.get(dName);
		} else {
			if (currObj.containsKey("use")) {
				retval=getObjDirective(allObjects, currObj.get("use"), dName);
			}
		}

		return(retval);
	}

	boolean findList(String l, String f) {
		StringTokenizer tokens = new StringTokenizer(l,",");
		while (tokens.hasMoreTokens()) {
			String aToken = tokens.nextToken().trim();
			if (aToken.equals(f)) {
				return(true);
			}
		}
		return (false);
	}

	Hashtable<String,Hashtable<String,String>> parseNagObjects(String objName){
	 	//a nagios objects of type objName will be stored in a hashtable, with a String as a key,
		// and its own Hashtable as a value--that hashtable represents the properties of the object
		//for key, use the 'name' field if present, the 'objName_name' field if present, or generate something.
		Hashtable<String, Hashtable<String,String>> tempTable = new Hashtable<String,Hashtable<String,String>>();
		StringBuffer config = new StringBuffer();
		if (MONITORING_VERSION == 1) {
			config = parseConfig(MONITORING_CONFIG);
		} else {
                        config = parseConfig(OBJECT_CACHE);
                        //config = parseObjectCache();
                }
			
		String line;
		BufferedReader br = new BufferedReader(new StringReader(config.toString()));
		try {
			while (((line = br.readLine()) != null)){
				if ( line.equals("")) { continue; };
                                StringTokenizer tokens = new StringTokenizer(line);
				if (tokens.nextToken().equals("define") && tokens.nextToken().equals(objName)) {
					Hashtable<String,String> nagObj = new Hashtable<String,String>();
					String name = "";
					line = br.readLine();
					while ((line != null) && ( parseNagLine(line,nagObj) )) {
						line = br.readLine();
					}
					if (nagObj.containsKey("name")) {
						name=(String) nagObj.get("name");
					} else if (nagObj.containsKey(objName+"_name")) {
						name=(String) nagObj.get(objName+"_name");
					} else {
						name="perf2rrd+" + objName + (tempTable.size());
					}

					tempTable.put(name, nagObj);
				}
			}
		} catch (Exception ex) {
			System.out.println("Problem parsing Nagios Objects: " + ex);
		}

		return(tempTable);

	}
        
        StringBuffer parseObjectCache() {
            StringBuffer config = new StringBuffer("");
            
            try {
                FileReader cache = new FileReader(OBJECT_CACHE);
                BufferedReader br = new BufferedReader(cache);
                String line;
                while ( (line = br.readLine()) != null) {
                    config.append(line + "\n" );
                }
            } catch (Exception ex) {
                System.out.println("Problem reading Object Cache file.");
                return (config);
            }
            
            return (config);
            
        }
        
	boolean parseNagLine(String line, Hashtable<String,String> obj) {
		//will return false when at end of object (key is closing '}' )
		//first token is key, rest of the line contructs the value
		StringTokenizer tokens = new StringTokenizer(line);
		String key="", value;

		if (tokens.hasMoreTokens()) {
			try {
				key=tokens.nextToken();
				if (key.equals("}") || key.equals("")) {
					return(false);
				}
			} catch (Exception ex) {
			}
		}

		value=line.substring(line.indexOf(key) + key.length()).trim();
		obj.put(key,value.toString());

		return(true);
	}

        String stringFix(String fix) {
            //prepare a string to be part of a filename
            String [] search = { "/", " " };
            String [] replace = { "-", "_" };
            
            for (int i = 0; i < search.length ; i++) {
                fix=fix.replaceAll(search[i], replace[i]);
            }
            
            return(fix);

    }
        
//sub-classes
	class PerfItem {
		//value types
		public static final int GAUGE=0;
		public static final int COUNTER=1;
		public static final int PERCENT=2;

                String perfData="";
		String itemName="";
		String itemVal=""; //this is a number that should be kept as a string..for the rrd update later
		Double itemMax=null;
		Double itemMin=null;
		int    itemType=GAUGE;
                Pattern NVT=Pattern.compile("([^=]*)=([\\d-.]*)(.*)");
                
		PerfItem() {
			super();
		}

		PerfItem(String s) {
			//submit string in form of  itemName=itemVal(itemType);warn,crit,min,max
			//warn and crit are not needed
                    
                        perfData=s;  //save for advanced parsing later.
			StringTokenizer tokens= new StringTokenizer(s,";", true);
			String aToken;
			try {
				aToken=tokens.nextToken();
				if (!(aToken.equals(";"))) {	//if it was not null
                                        Matcher m=NVT.matcher(aToken);
                                        if (m.find()) {
                                            itemName=m.group(1);
                                            itemVal=(m.group(2));
                                        }
				}
			}  catch (Exception ex) {
				// leave whatever else to defaults, when tokens run out code should jump here.
			}


		}
                
                void parseFull() {
                    // Parsing all of the element is only necessary when creating a new RRD
                    
			StringTokenizer tokens= new StringTokenizer(perfData,";", true);
			String aToken;
                        String tempType;
			try {
				aToken=tokens.nextToken();
				if (!(aToken.equals(";"))) {	//if it was not null
					Matcher m=NVT.matcher(aToken);
                                        if (m.find()) {
                                            itemName=m.group(1);
                                            itemVal=(m.group(2));
                                            if (! m.group(3).equals("") ) {
                                                tempType=m.group(3);
                                                if (tempType.equals("%")) {
                                                    itemType=PERCENT;
                                                } else if (tempType.equals("c")) {
                                                    itemType=COUNTER;
                                                } else {
                                                    itemType=GAUGE;
                                                }
                                              
                                            }
                                        }
                                    
					tokens.nextToken();   //this should read out the semi
				}
				if (itemType==PERCENT) {
					itemMin=new Double(0.0);
					itemMax=new Double(100.0);
				}

				if (!tokens.nextToken().equals(";")) {  //skip warn if present,and next semi
					tokens.nextToken();
				}
				if (!tokens.nextToken().equals(";")) {  //skip crit if present,and next semi
					tokens.nextToken();
				}
				aToken=tokens.nextToken();
				if (!(aToken.equals(";"))) {
					itemMin=new Double(Double.parseDouble(aToken));
					tokens.nextToken(); //this should read out the semi
				}
				aToken=tokens.nextToken();
				if (!(aToken.equals(";"))) {
					itemMax=new Double(Double.parseDouble(aToken));
				}
			}  catch (Exception ex) {
				// leave whatever else to defaults, when tokens run out code should jump here.
			}
                    
                }

		
		public String toString() {
			return("Name: " + itemName + " Val: " + itemVal + " Min:" + itemMin + " Max: " + itemMax + " Type: " + itemType);
		}

		public String getItemName() {
			return itemName;
		}

		public int getItemType() {
			return itemType;
		}

		public Double getItemMin() {
			return itemMin;
		}

		public Double getItemMax() {
			return itemMax;
		}
		public String getItemValue() {
			return itemVal;
		}

	}
  

}
